/*
 * Copyright (c) 2017 Samsung Electronics Co., Ltd All Rights Reserved
 *
 *    Licensed under the Apache License, Version 2.0 (the "License");
 *    you may not use this file except in compliance with the License.
 *    You may obtain a copy of the License at
 *
 *        http://www.apache.org/licenses/LICENSE-2.0
 *
 *    Unless required by applicable law or agreed to in writing, software
 *    distributed under the License is distributed on an "AS IS" BASIS,
 *    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *    See the License for the specific language governing permissions and
 *    limitations under the License.
 */


var utils_ = xwalk.utils;
var type_ = utils_.type;
var converter_ = utils_.converter;
var validator_ = utils_.validator;
var types_ = validator_.Types;
var native_ = new xwalk.utils.NativeManager(extension);

function ListenerManager(native, listenerName, handle) {
    this.listeners = {};
    this.nextId = 1;
    this.nativeSet = false;
    this.native = native;
    this.listenerName = listenerName;
    this.handle = handle || function() {};
}

ListenerManager.prototype.addListener = function(callback, nativeCall, data) {
    var id = this.nextId;
    if (!this.nativeSet) {
        this.native.addListener(this.listenerName, function(msg) {
            for (var watchId in this.listeners) {
                if (this.listeners.hasOwnProperty(watchId)) {
                    this.handle(msg, this.listeners[watchId], watchId);
                }
            }
        }.bind(this));
        var result = this.native.callSync(nativeCall, data || {});
        if (this.native.isFailure(result)) {
            throw this.native.getErrorObject(result);
        }
        this.nativeSet = true;
    }

    this.listeners[id] = callback;
    ++this.nextId;

    return id;
};

ListenerManager.prototype.removeListener = function(watchId, nativeCall) {
    if (this.listeners.hasOwnProperty(watchId)) {
        delete this.listeners[watchId];
    }

    if (this.nativeSet && type_.isEmptyObject(this.listeners)) {
        this.native.callSync(nativeCall);
        this.native.removeListener(this.listenerName);
        this.nativeSet = false;
    }
};


var TeecLoginMethod = {
    PUBLIC: 'PUBLIC',
    USER: 'USER',
    GROUP: 'GROUP',
    APPLICATION: 'APPLICATION'
};
var TeecValueType = {
    INPUT: 'INPUT',
    OUTPUT: 'OUTPUT',
    INOUT: 'INOUT'
};
var TeecTempMemoryType = {
    INPUT: 'INPUT',
    OUTPUT: 'OUTPUT',
    INOUT: 'INOUT'
};
var TeecRegisteredMemoryType = {
    WHOLE: 'WHOLE',
    PARTIAL_INPUT: 'PARTIAL_INPUT',
    PARTIAL_OUTPUT: 'PARTIAL_OUTPUT',
    PARTIAL_INOUT: 'PARTIAL_INOUT'
};
var TeecSharedMemoryFlags = {
    INPUT: 'INPUT',
    OUTPUT: 'OUTPUT',
    INOUT: 'INOUT'
};
var TEEC_MAX_OPERATION_LENGTH = 4;
var TEEC_PARAMETER_TYPE_VALUE = 'VALUE';
var TEEC_PARAMETER_TYPE_MEMREF = 'MEMREF';
var TEEC_PARAMETER_TYPE_TMPREF = 'TMPREF';


function TeecParameter() {
}



function LibTeecManager() {
}

LibTeecManager.prototype.getContext = function() {
    var args = validator_.validateArgs(arguments, [
        {name: 'name', type: types_.STRING, optional: true, nullable: true}
    ]);

    var data = {
        name: args.name
    };

    var result = native_.callSync('LibTeecManager_getContext', data);

    if (native_.isFailure(result)) {
        throw native_.getErrorObject(result);
    }

    return new TeecContext(native_.getResultObject(result));
};



function TeecSession(contextId, sessionId) {
    this.contextId = contextId;
    this.sessionId = sessionId;
}

TeecSession.prototype.close = function() {
    var data = {
        contextId: this.contextId,
        sessionId: this.sessionId
    };

    var result = native_.callSync('TeecSession_close', data);

    if (native_.isFailure(result)) {
        throw native_.getErrorObject(result);
    }
};

TeecSession.prototype.invokeCommand =
    function(cmd, params, successCallback, errorCallback) {
    var args = validator_.validateArgs(arguments, [
        {name: 'cmd', type: types_.LONG},
        {name: 'params', type: types_.ARRAY, values: TeecParameter,
            nullable: true},
        {name: 'successCallback', type: types_.FUNCTION},
        {name: 'errorCallback', type: types_.FUNCTION, optional: true,
            nullable: true}
    ]);

    if (params.length > TEEC_MAX_OPERATION_LENGTH) {
        throw new WebAPIException(WebAPIException.INVALID_VALUES_ERR,
            'Input array params has too high length');
    }

    var data = {
        contextId: this.contextId,
        sessionId: this.sessionId,
        cmd: args.cmd,
        params: args.params,
        successCallback: args.successCallback,
        errorCallback: args.errorCallback
    };

    var callback = function(result) {
        if (native_.isFailure(result)) {
            native_.callIfPossible(args.errorCallback,
                                   native_.getErrorObject(result));
            return;
        }

        var retArgs = native_.getResultObject(result);

        // update parameters after call
        // to avoid losing TeecSharedMemory's functions, we copy MEMREF
        // attributes via a function, instead of a typical substitution
        for (var i = 0; i < params.length; ++i) {
            switch (params[i].type) {
            case TEEC_PARAMETER_TYPE_VALUE:
                params[i] = retArgs.params[i];
                break;
            case TEEC_PARAMETER_TYPE_MEMREF:
                params[i].copyParams(retArgs.params[i]);
                break;
            case TEEC_PARAMETER_TYPE_TMPREF:
                params[i] = retArgs.params[i];
                break;
            default:
                throw new WebAPIException(WebAPIException.INVALID_VALUES_ERR,
                    'Received incorrect param object from plugin');
            };
        }

        native_.callIfPossible(args.successCallback, retArgs.cmd, params);
    };

    var result = native_.call('TeecSession_invokeCommand', data, callback);

    if (native_.isFailure(result)) {
        throw native_.getErrorObject(result);
    }

    return native_.getResultObject(result);
};



function TeecContext(id) {
    // constructor of TeecContext
    this.contextId = id;
}

TeecContext.prototype.openSession =
    function(taUUID, loginMethod, connectionData, params, successCallback,
             errorCallback) {
    var args = validator_.validateArgs(arguments, [
        {name: 'taUUID', type: types_.STRING},
        {name: 'loginMethod', type: types_.ENUM,
            values: ['PUBLIC', 'USER', 'GROUP', 'APPLICATION']},
        {name: 'connectionData', type: types_.UNSIGNED_LONG, nullable: true},
        {name: 'params', type: types_.ARRAY, values: TeecParameter,
            nullable: true},
        {name: 'successCallback', type: types_.FUNCTION},
        {name: 'errorCallback', type: types_.FUNCTION, optional: true,
            nullable: true}
    ]);

    var data = {
        contextId: this.contextId,
        taUUID: args.taUUID,
        loginMethod: args.loginMethod,
        connectionData: args.connectionData,
        params: args.params
    };

    var cId = this.contextId;
    var callback = function(result) {
        if (native_.isFailure(result)) {
            native_.callIfPossible(args.errorCallback, native_.getErrorObject(result));
            return;
        }

        var session = new TeecSession(cId, native_.getResultObject(result));
        native_.callIfPossible(args.successCallback, session);
    };

    var result = native_.call('TeecContext_openSession', data, callback);

    if (native_.isFailure(result)) {
        throw native_.getErrorObject(result);
    }

    return native_.getResultObject(result);
};

TeecContext.prototype.revokeCommand = function(id) {
    var args = validator_.validateArgs(arguments, [
        {name: 'id', type: types_.LONG}
    ]);

    var data = {
        id: args.id
    };

    var result = native_.callSync('TeecContext_revokeCommand', data);

    if (native_.isFailure(result)) {
        throw native_.getErrorObject(result);
    }
};

TeecContext.prototype.allocateSharedMemory = function(size, flags) {
    var args = validator_.validateArgs(arguments, [
        {name: 'size', type: types_.LONG},
        {name: 'flags', type: types_.ENUM, values: ['INPUT', 'OUTPUT', 'INOUT']}
    ]);

    var data = {
        contextId: this.contextId,
        size: args.size,
        flags: args.flags
    };

    var result = native_.callSync('TeecContext_allocateSharedMemory', data);

    if (native_.isFailure(result)) {
        throw native_.getErrorObject(result);
    }

    return new TeecSharedMemory(this.contextId, native_.getResultObject(result));
};

TeecContext.prototype.registerSharedMemory = function(addr, size, flags) {
    var args = validator_.validateArgs(arguments, [
        {name: 'addr', type: types_.LONG_LONG},
        {name: 'size', type: types_.LONG},
        {name: 'flags', type: types_.ENUM, values: ['INPUT', 'OUTPUT', 'INOUT']}
    ]);

    var data = {
        addr: args.addr,
        size: args.size,
        flags: args.flags
    };

    var result = native_.callSync('TeecContext_registerSharedMemory', data);

    if (native_.isFailure(result)) {
        throw native_.getErrorObject(result);
    }

    return new TeecSharedMemory(this.contextId, native_.getResultObject(result));
};

TeecContext.prototype.releaseSharedMemory = function(shm) {
    var args = validator_.validateArgs(arguments, [
        {name: 'shm', type: types_.PLATFORM_OBJECT, values: tizen.TeecSharedMemory}
    ]);

    var data = {
        contextId: this.contextId,
        memId: args.shm.memId
    };

    var result = native_.callSync('TeecContext_releaseSharedMemory', data);

    if (native_.isFailure(result)) {
        throw native_.getErrorObject(result);
    }
};



function TeecSharedMemory(contextId, memId) {
    // constructor of TeecSharedMemory
    this.contextId = contextId;
    this.memId = memId;
}

TeecSharedMemory.prototype.setData = function(data, offset) {
    var args = validator_.validateArgs(arguments, [
        {name: 'data', type: types_.ARRAY, values: types_.BYTE },
        {name: 'offset', type: types_.LONG_LONG }
    ]);

    var packed = {
        contextId: this.contextId,
        memId: this.memId,
        data: args.data,
        offset: args.offset
    };

    var result = native_.callSync('TeecSharedMemory_setData', packed);

    if (native_.isFailure(result)) {
        throw native_.getErrorObject(result);
    }
};

TeecSharedMemory.prototype.getData = function(data, offset) {
    var args = validator_.validateArgs(arguments, [
        {name: 'data', type: types_.ARRAY, values: types_.BYTE },
        {name: 'offset', type: types_.LONG_LONG}
    ]);

    var packed = {
        contextId: this.contextId,
        memId: this.memId,
        data: args.data,
        offset: args.offset
    };

    var result = native_.callSync('TeecSharedMemory_getData', packed);

    if (native_.isFailure(result)) {
        throw native_.getErrorObject(result);
    }

    var out = native_.getResultObject(result);
    for (var i = 0; i < data.length; ++i)
        data[i] = out[i];
};

TeecSharedMemory.prototype.copyParams = function(src) {
    this.contextId = src.contextId;
    this.memId = src.memId;
};



function TeecValue(a, b, flag) {
    // constructor of TeecValue
    validator_.isConstructorCall(this, TeecValue);
    var args = validator_.validateArgs(arguments, [
        {name: 'a', type: types_.UNSIGNED_LONG},
        {name: 'b', type: types_.UNSIGNED_LONG},
        {name: 'flag', type: types_.ENUM, values: ['INPUT', 'OUTPUT', 'INOUT']}
    ]);

    this.type = TEEC_PARAMETER_TYPE_VALUE;
    this.flag = args.flag;
    this.a = args.a;
    this.b = args.b;
}

TeecValue.prototype = new TeecParameter();
TeecValue.prototype.constructor = TeecValue;

function TeecRegisteredMemory(memory, offset, size, flag) {
    // constructor of TeecRegisteredMemory
    validator_.isConstructorCall(this, TeecRegisteredMemory);
    var args = validator_.validateArgs(arguments, [
        {name: 'memory', type: types_.PLATFORM_OBJECT, values: TeecSharedMemory },
        {name: 'offset', type: types_.UNSIGNED_LONG_LONG },
        {name: 'size', type: types_.UNSIGNED_LONG_LONG },
        {name: 'flag', type: types_.ENUM, values: ['WHOLE', 'PARTIAL_INPUT', 'PARTIAL_OUTPUT', 'PARTIAL_INOUT']}
    ]);

    this.type = TEEC_PARAMETER_TYPE_MEMREF;
    this.flag = args.flag;
    this.shm = memory;
    this.size = args.size;
    this.offset = args.offset;
}

TeecRegisteredMemory.prototype = new TeecParameter();
TeecRegisteredMemory.prototype.constructor = TeecRegisteredMemory;
TeecRegisteredMemory.prototype.copyParams = function(src) {
    this.type = src.type;
    this.flag = src.flag;
    this.size = src.size;
    this.offset = src.offset;
    this.shm.copyParams(src.shm);
};


function TeecTempMemory(mem, flag) {
    // constructor of TeecTempMemory
    validator_.isConstructorCall(this, TeecTempMemory);
    var args = validator_.validateArgs(arguments, [
        {name: 'mem', type: types_.ARRAY, values: types_.BYTE },
        {name: 'flag', type: types_.ENUM, values: ['INPUT', 'OUTPUT', 'INOUT']}
    ]);

    this.type = TEEC_PARAMETER_TYPE_TMPREF;
    this.mem = args.mem;
    this.flag = args.flag;
}

TeecTempMemory.prototype = new TeecParameter();
TeecTempMemory.prototype.constructor = TeecTempMemory;



exports = new LibTeecManager();
tizen.TeecRegisteredMemory = TeecRegisteredMemory;
tizen.TeecTempMemory = TeecTempMemory;
tizen.TeecValue = TeecValue;
