/**
 * Copyright (c) 2017 Samsung Electronics Co., Ltd All Rights Reserved
 *
 *    Licensed under the Apache License, Version 2.0 (the "License");
 *    you may not use this file except in compliance with the License.
 *    You may obtain a copy of the License at
 *
 *        http://www.apache.org/licenses/LICENSE-2.0
 *
 *    Unless required by applicable law or agreed to in writing, software
 *    distributed under the License is distributed on an "AS IS" BASIS,
 *    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *    See the License for the specific language governing permissions and
 *    limitations under the License.
 */
/**
 * @file
 * @author Lukasz Kostyra (l.kostyra@samsung.com)
 * @brief  Libteec instance definitions
 */

#include "libteec_instance.h"
#include "TeecManager.h"
#include "TeecContext.h"
#include "TeecTranslations.h"
#include "TeecTempMemoryAllocator.h"

#include <functional>

#include <common/picojson.h>
#include <common/logger.h>
#include <common/platform_exception.h>
#include <common/task-queue.h>
#include <common/converter.h>
#include <common/tools.h>


namespace {

// The privilege required in Libteec API
const std::string kPrivilegeLibteec = "http://tizen.org/privilege/tee.client";

} // namespace

namespace extension {
namespace libteec {

using namespace common;
using namespace extension::libteec;

LibteecInstance::LibteecInstance() {
    using namespace std::placeholders;
    #define REGISTER_SYNC(c,x) \
        RegisterSyncHandler(c, std::bind(&LibteecInstance::x, this, _1, _2));
    REGISTER_SYNC("LibTeecManager_getContext", LibTeecManagerGetContext);
    REGISTER_SYNC("TeecSharedMemory_getData", TeecSharedMemoryGetData);
    REGISTER_SYNC("TeecSharedMemory_setData", TeecSharedMemorySetData);
    REGISTER_SYNC("TeecContext_releaseSharedMemory", TeecContextReleaseSharedMemory);
    REGISTER_SYNC("TeecContext_registerSharedMemory", TeecContextRegisterSharedMemory);
    REGISTER_SYNC("TeecContext_allocateSharedMemory", TeecContextAllocateSharedMemory);
    REGISTER_SYNC("TeecContext_revokeCommand", TeecContextRevokeCommand);
    REGISTER_SYNC("TeecSession_close", TeecSessionClose);
    #undef REGISTER_SYNC

    #define REGISTER_ASYNC(c,x) \
        RegisterSyncHandler(c, std::bind(&LibteecInstance::x, this, _1, _2));
    REGISTER_ASYNC("TeecContext_openSession", TeecContextOpenSession);
    REGISTER_ASYNC("TeecSession_invokeCommand", TeecSessionInvokeCommand);
    #undef REGISTER_ASYNC
}

#define CHECK_EXIST(args, name, out) \
        if (!args.contains(name)) {\
            ReportError(TypeMismatchException(name" is required argument"), out);\
            return;\
        }

void LibteecInstance::LibTeecManagerGetContext(const picojson::value& args, picojson::object& out) const {
    CHECK_PRIVILEGE_ACCESS(kPrivilegeLibteec, &out);

    const picojson::value& name = args.get("name");
    std::string contextName;

    if (name.is<std::string>()) {
        contextName = name.get<std::string>();
    }

    try {
        TeecManager::Instance().GetContext(contextName);
        ReportSuccess(picojson::value(contextName), out);
    } catch (const PlatformException& e) {
        ReportError(e, out);
    }
}

void LibteecInstance::TeecContextOpenSession(const picojson::value& args, picojson::object& out) {
    CHECK_PRIVILEGE_ACCESS(kPrivilegeLibteec, &out);
    CHECK_EXIST(args, "callbackId", out)
    CHECK_EXIST(args, "contextId", out)
    CHECK_EXIST(args, "taUUID", out)
    CHECK_EXIST(args, "loginMethod", out)

    double callbackId = args.get("callbackId").get<double>();
    std::string contextId = args.get("contextId").get<std::string>();
    std::string uuid = args.get("taUUID").get<std::string>();
    const uint32_t loginMethod = translations::StringToTeecLoginMethod(args.get("loginMethod").get<std::string>());

    TeecContextPtr context;
    uint32_t connectionData = 0;
    uint32_t* connectionDataPtr = nullptr;
    uint32_t opId = 0;

    try {
        context = TeecManager::Instance().GetContext(contextId);

        if (args.get("connectionData").is<double>()) {
            connectionData = static_cast<uint32_t>(args.get("connectionData").get<double>());
            connectionDataPtr = &connectionData;
        }

        opId = TeecManager::Instance().CreateOperation();
    } catch (const PlatformException& e) {
        ReportError(e, out);
    }

    auto openSession = [=](const std::shared_ptr<picojson::value>& response) -> void {
        try {
            TeecTempMemoryAllocator allocator;

            TEEC_Operation op = TeecManager::Instance().GetOperation(opId);
            op.paramTypes = TEEC_PARAM_TYPES(TEEC_NONE, TEEC_NONE, TEEC_NONE, TEEC_NONE);
            if (args.get("params").is<picojson::array>()) {
                op = translations::OperationFromJSON(args.get("params"), allocator);
            }

            std::string sid = context->OpenSession(uuid, loginMethod, connectionDataPtr, &op, nullptr);
            TeecManager::Instance().RemoveOperation(opId);
            ReportSuccess(picojson::value(sid), response->get<picojson::object>());
        } catch (const PlatformException& e) {
            ReportError(e, response->get<picojson::object>());
        }
    };

    auto openSessionResponse =
        [callbackId, this](const std::shared_ptr<picojson::value>& response) -> void {
        picojson::object& obj = response->get<picojson::object>();
        obj.insert(std::make_pair("callbackId", picojson::value(callbackId)));
        Instance::PostMessage(this, response->serialize().c_str());
    };

    auto data = std::shared_ptr<picojson::value>(new picojson::value(picojson::object()));

    TaskQueue::GetInstance().Queue<picojson::value>(openSession, openSessionResponse, data);

    ReportSuccess(picojson::value(static_cast<double>(opId)), out);
}

void LibteecInstance::TeecContextRevokeCommand(const picojson::value& args, picojson::object& out) const {
    CHECK_PRIVILEGE_ACCESS(kPrivilegeLibteec, &out);
    CHECK_EXIST(args, "id", out);

    try {
        uint32_t id = static_cast<uint32_t>(args.get("id").get<double>());
        TEEC_Operation op = TeecManager::Instance().GetOperation(id);
        TEEC_RequestCancellation(&op);
        ReportSuccess(out);
    } catch (const PlatformException& e) {
        ReportError(e, out);
    }
}

void LibteecInstance::TeecContextAllocateSharedMemory(const picojson::value& args, picojson::object& out) const {
    CHECK_PRIVILEGE_ACCESS(kPrivilegeLibteec, &out);
    CHECK_EXIST(args, "contextId", out);
    CHECK_EXIST(args, "size", out);
    CHECK_EXIST(args, "flags", out);

    size_t size = static_cast<size_t>(args.get("size").get<double>());
    uint32_t flags = translations::SharedMemoryFlagsFromJSON(args.get("flags"));
    std::string contextId = args.get("contextId").get<std::string>();

    try {
        TeecContextPtr context = TeecManager::Instance().GetContext(contextId);
        std::string memId = context->CreateSharedMemory(size, flags);
        ReportSuccess(picojson::value(memId), out);
    } catch (const PlatformException& e) {
        ReportError(e, out);
    }
}

void LibteecInstance::TeecContextRegisterSharedMemory(const picojson::value& args, picojson::object& out) const {
    CHECK_PRIVILEGE_ACCESS(kPrivilegeLibteec, &out);
    CHECK_EXIST(args, "contextId", out);
    CHECK_EXIST(args, "buffer", out);
    CHECK_EXIST(args, "size", out);
    CHECK_EXIST(args, "flags", out);

    void* buffer = reinterpret_cast<void*>(static_cast<size_t>(args.get("buffer").get<double>()));
    size_t size = static_cast<size_t>(args.get("size").get<double>());
    uint32_t flags = static_cast<uint32_t>(args.get("flags").get<double>());
    std::string contextId = args.get("contextId").get<std::string>();

    try {
        TeecContextPtr context = TeecManager::Instance().GetContext(contextId);
        std::string memId = context->CreateSharedMemory(buffer, size, flags);
        ReportSuccess(picojson::value(memId), out);
    } catch (const PlatformException& e) {
        ReportError(e, out);
    }
}

void LibteecInstance::TeecContextReleaseSharedMemory(const picojson::value& args, picojson::object& out) const {
    CHECK_PRIVILEGE_ACCESS(kPrivilegeLibteec, &out);
    CHECK_EXIST(args, "contextId", out);
    CHECK_EXIST(args, "memId", out);

    std::string contextId = args.get("contextId").get<std::string>();
    std::string memId = args.get("memId").get<std::string>();

    try {
        TeecContextPtr context = TeecManager::Instance().GetContext(contextId);
        context->ReleaseSharedMemory(memId);
    } catch (const PlatformException& e) {
        ReportError(e, out);
    }
}

void LibteecInstance::TeecSharedMemorySetData(const picojson::value& args, picojson::object& out) const {
    CHECK_PRIVILEGE_ACCESS(kPrivilegeLibteec, &out);
    CHECK_EXIST(args, "data", out);
    CHECK_EXIST(args, "offset", out);

    std::string contextId = args.get("contextId").get<std::string>();
    std::string memId = args.get("memId").get<std::string>();
    DataBuffer data = translations::DataBufferFromJSON(args.get("data"));
    size_t offset = static_cast<size_t>(args.get("offset").get<double>());

    try {
        TeecContextPtr context = TeecManager::Instance().GetContext(contextId);
        TeecSharedMemoryPtr shmem = context->GetSharedMemory(memId);
        shmem->SetData(data, offset);
        ReportSuccess(out);
    } catch (const PlatformException& e) {
        ReportError(e, out);
    }
}

void LibteecInstance::TeecSharedMemoryGetData(const picojson::value& args, picojson::object& out) const {
    CHECK_PRIVILEGE_ACCESS(kPrivilegeLibteec, &out);
    CHECK_EXIST(args, "data", out);
    CHECK_EXIST(args, "offset", out);

    std::string contextId = args.get("contextId").get<std::string>();
    std::string memId = args.get("memId").get<std::string>();
    DataBuffer data = translations::DataBufferFromJSON(args.get("data"));
    size_t offset = static_cast<size_t>(args.get("offset").get<double>());

    try {
        TeecContextPtr context = TeecManager::Instance().GetContext(contextId);
        TeecSharedMemoryPtr shmem = context->GetSharedMemory(memId);
        shmem->GetData(data, offset);
        ReportSuccess(translations::JSONFromDataBuffer(data), out);
    } catch (const PlatformException& e) {
        ReportError(e, out);
    }
}

void LibteecInstance::TeecSessionClose(const picojson::value& args, picojson::object& out) const {
    CHECK_PRIVILEGE_ACCESS(kPrivilegeLibteec, &out);
    CHECK_EXIST(args, "contextId", out);
    CHECK_EXIST(args, "sessionId", out);

    const std::string contextId = args.get("contextId").get<std::string>();
    const std::string sessionId = args.get("sessionId").get<std::string>();

    try {
        TeecContextPtr context = TeecManager::Instance().GetContext(contextId);
        context->CloseSession(sessionId);
        ReportSuccess(out);
    } catch (const PlatformException& e) {
        ReportError(e, out);
    }
}

void LibteecInstance::TeecSessionInvokeCommand(const picojson::value& args, picojson::object& out) {
    CHECK_PRIVILEGE_ACCESS(kPrivilegeLibteec, &out);
    CHECK_EXIST(args, "callbackId", out);
    CHECK_EXIST(args, "contextId", out);
    CHECK_EXIST(args, "sessionId", out);
    CHECK_EXIST(args, "cmd", out);

    const double callbackId = args.get("callbackId").get<double>();
    const std::string contextId = args.get("contextId").get<std::string>();
    const std::string sessionId = args.get("sessionId").get<std::string>();
    const uint32_t cmd = static_cast<uint32_t>(args.get("cmd").get<double>());

    TeecContextPtr context;
    TeecSessionPtr session;
    uint32_t opId = 0;

    try {
        context = TeecManager::Instance().GetContext(contextId);
        session = context->GetSession(sessionId);
        opId = TeecManager::Instance().CreateOperation();
    } catch (const PlatformException& e) {
        ReportError(e, out);
    }

    auto invoke = [=](const std::shared_ptr<picojson::value>& response) -> void {
        try {
            TeecTempMemoryAllocator allocator;
            TEEC_Operation op = TeecManager::Instance().GetOperation(opId);
            op.paramTypes = TEEC_PARAM_TYPES(TEEC_NONE, TEEC_NONE, TEEC_NONE, TEEC_NONE);
            if (args.get("params").is<picojson::array>()) {
                op = translations::OperationFromJSON(args.get("params"), allocator);
            }

            session->InvokeCommand(cmd, &op, NULL);
            TeecManager::Instance().RemoveOperation(opId);

            picojson::object result;
            result.insert(std::make_pair("cmd", picojson::value(static_cast<double>(cmd))));
            result.insert(std::make_pair("params", translations::UpdateJSONParams(args.get("params"), op)));
            ReportSuccess(picojson::value(result), response->get<picojson::object>());
        } catch (const PlatformException& e) {
            ReportError(e, response->get<picojson::object>());
        }
    };

    auto invokeResponse =
        [callbackId, this](const std::shared_ptr<picojson::value>& response) -> void {
        picojson::object& obj = response->get<picojson::object>();
        obj.insert(std::make_pair("callbackId", picojson::value(callbackId)));
        Instance::PostMessage(this, response->serialize().c_str());
    };

    auto data = std::shared_ptr<picojson::value>(new picojson::value(picojson::object()));

    TaskQueue::GetInstance().Queue<picojson::value>(invoke, invokeResponse, data);
    ReportSuccess(picojson::value(static_cast<double>(opId)), out);
}

#undef CHECK_EXIST

} // namespace libteec
} // namespace extension